"""Test https://github.com/Textualize/textual/issues/4248"""

from textual.app import App, ComposeResult
from textual.widgets import Label


async def test_issue_4248() -> None:
    """Various forms of click parameters should be fine."""

    bumps = 0

    class ActionApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Label("[@click]click me and crash[/]", id="nothing")
            yield Label("[@click=]click me and crash[/]", id="no-params")
            yield Label("[@click=()]click me and crash[/]", id="empty-params")
            yield Label("[@click=foobar]click me[/]", id="unknown-sans-parens")
            yield Label("[@click=foobar()]click me[/]", id="unknown-with-parens")
            yield Label("[@click=app.bump]click me[/]", id="known-sans-parens")
            yield Label("[@click=app.bump()]click me[/]", id="known-empty-parens")
            yield Label("[@click=app.bump(100)]click me[/]", id="known-with-param")

        def action_bump(self, by_value: int = 1) -> None:
            nonlocal bumps
            bumps += by_value

    app = ActionApp()
    async with app.run_test() as pilot:
        assert bumps == 0
        await pilot.click("#nothing")
        assert bumps == 0
        await pilot.click("#no-params")
        assert bumps == 0
        await pilot.click("#empty-params")
        assert bumps == 0
        await pilot.click("#unknown-sans-parens")
        assert bumps == 0
        await pilot.click("#unknown-with-parens")
        assert bumps == 0
        await pilot.click("#known-sans-parens")
        assert bumps == 1
        await pilot.click("#known-empty-parens")
        assert bumps == 2
        await pilot.click("#known-with-param")
        assert bumps == 102



from textual.app import App, ComposeResult
from textual.widgets import DirectoryTree


class DirectoryTreeApp(App):
    def compose(self) -> ComposeResult:
        yield DirectoryTree("./")


if __name__ == "__main__":
    app = DirectoryTreeApp()
    app.run()



"""

Builtin events sent by Textual.

Events may be marked as "Bubbles" and "Verbose".
See the [events guide](/guide/events/#bubbling) for an explanation of bubbling.
Verbose events are excluded from the textual console, unless you explicitly request them with the `-v` switch as follows:

```
textual console -v
```
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Type, TypeVar

import rich.repr
from rich.style import Style

from textual._types import CallbackType
from textual.geometry import Offset, Size
from textual.keys import _get_key_aliases
from textual.message import Message

MouseEventT = TypeVar("MouseEventT", bound="MouseEvent")

if TYPE_CHECKING:
    from textual.dom import DOMNode
    from textual.timer import Timer as TimerClass
    from textual.timer import TimerCallback
    from textual.widget import Widget


@rich.repr.auto
class Event(Message):
    """The base class for all events."""


@rich.repr.auto
class Callback(Event, bubble=False, verbose=True):
    """Sent by Textual to invoke a callback
    (see [call_next][textual.message_pump.MessagePump.call_next] and
    [call_later][textual.message_pump.MessagePump.call_later]).
    """

    def __init__(self, callback: CallbackType) -> None:
        self.callback = callback
        super().__init__()

    def __rich_repr__(self) -> rich.repr.Result:
        yield "callback", self.callback


@dataclass
class CursorPosition(Event, bubble=False):
    """Internal event used to retrieve the terminal's cursor position."""

    x: int
    y: int


class Load(Event, bubble=False):
    """
    Sent when the App is running but *before* the terminal is in application mode.

    Use this event to run any setup that doesn't require any visuals such as loading
    configuration and binding keys.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Idle(Event, bubble=False):
    """Sent when there are no more items in the message queue.

    This is a pseudo-event in that it is created by the Textual system and doesn't go
    through the usual message queue.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Action(Event):
    __slots__ = ["action"]

    def __init__(self, action: str) -> None:
        super().__init__()
        self.action = action

    def __rich_repr__(self) -> rich.repr.Result:
        yield "action", self.action


class Resize(Event, bubble=False):
    """Sent when the app or widget has been resized.

    - [ ] Bubbles
    - [ ] Verbose

    Args:
        size: The new size of the Widget.
        virtual_size: The virtual size (scrollable size) of the Widget.
        container_size: The size of the Widget's container widget.
    """

    __slots__ = ["size", "virtual_size", "container_size"]

    def __init__(
        self,
        size: Size,
        virtual_size: Size,
        container_size: Size | None = None,
    ) -> None:
        self.size = size
        """The new size of the Widget."""
        self.virtual_size = virtual_size
        """The virtual size (scrollable size) of the Widget."""
        self.container_size = size if container_size is None else container_size
        """The size of the Widget's container widget."""
        super().__init__()

    def can_replace(self, message: "Message") -> bool:
        return isinstance(message, Resize)

    def __rich_repr__(self) -> rich.repr.Result:
        yield "size", self.size
        yield "virtual_size", self.virtual_size
        yield "container_size", self.container_size, self.size


class Compose(Event, bubble=False, verbose=True):
    """Sent to a widget to request it to compose and mount children.

    This event is used internally by Textual.
    You won't typically need to explicitly handle it,

    - [ ] Bubbles
    - [X] Verbose
    """


class Mount(Event, bubble=False, verbose=False):
    """Sent when a widget is *mounted* and may receive messages.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Unmount(Event, bubble=False, verbose=False):
    """Sent when a widget is unmounted and may no longer receive messages.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Show(Event, bubble=False):
    """Sent when a widget is first displayed.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Hide(Event, bubble=False):
    """Sent when a widget has been hidden.

    - [ ] Bubbles
    - [ ] Verbose

    Sent when any of the following conditions apply:

    - The widget is removed from the DOM.
    - The widget is no longer displayed because it has been scrolled or clipped from the terminal or its container.
    - The widget has its `display` attribute set to `False`.
    - The widget's `display` style is set to `"none"`.
    """


class Ready(Event, bubble=False):
    """Sent to the `App` when the DOM is ready and the first frame has been displayed.

    - [ ] Bubbles
    - [ ] Verbose
    """


@rich.repr.auto
class MouseCapture(Event, bubble=False):
    """Sent when the mouse has been captured.

    - [ ] Bubbles
    - [ ] Verbose

    When a mouse has been captured, all further mouse events will be sent to the capturing widget.

    Args:
        mouse_position: The position of the mouse when captured.
    """

    def __init__(self, mouse_position: Offset) -> None:
        super().__init__()
        self.mouse_position = mouse_position
        """The position of the mouse when captured."""

    def __rich_repr__(self) -> rich.repr.Result:
        yield None, self.mouse_position


@rich.repr.auto
class MouseRelease(Event, bubble=False):
    """Mouse has been released.

    - [ ] Bubbles
    - [ ] Verbose

    Args:
        mouse_position: The position of the mouse when released.
    """

    def __init__(self, mouse_position: Offset) -> None:
        super().__init__()
        self.mouse_position = mouse_position
        """The position of the mouse when released."""

    def __rich_repr__(self) -> rich.repr.Result:
        yield None, self.mouse_position


class InputEvent(Event):
    """Base class for input events."""


@rich.repr.auto
class Key(InputEvent):
    """Sent when the user hits a key on the keyboard.

    - [X] Bubbles
    - [ ] Verbose

    Args:
        key: The key that was pressed.
        character: A printable character or `None` if it is not printable.
    """

    __slots__ = ["key", "character", "aliases"]

    def __init__(self, key: str, character: str | None) -> None:
        super().__init__()
        self.key = key
        """The key that was pressed."""
        self.character = (
            (key if len(key) == 1 else None) if character is None else character
        )
        """A printable character or ``None`` if it is not printable."""
        self.aliases: list[str] = _get_key_aliases(key)
        """The aliases for the key, including the key itself."""

    def __rich_repr__(self) -> rich.repr.Result:
        yield "key", self.key
        yield "character", self.character
        yield "name", self.name
        yield "is_printable", self.is_printable
        yield "aliases", self.aliases, [self.key]

    @property
    def name(self) -> str:
        """Name of a key suitable for use as a Python identifier."""
        return _key_to_identifier(self.key).lower()

    @property
    def name_aliases(self) -> list[str]:
        """The corresponding name for every alias in `aliases` list."""
        return [_key_to_identifier(key) for key in self.aliases]

    @property
    def is_printable(self) -> bool:
        """Check if the key is printable (produces a unicode character).

        Returns:
            `True` if the key is printable.
        """
        return False if self.character is None else self.character.isprintable()


def _key_to_identifier(key: str) -> str:
    """Convert the key string to a name suitable for use as a Python identifier."""
    key_no_modifiers = key.split("+")[-1]
    if len(key_no_modifiers) == 1 and key_no_modifiers.isupper():
        if "+" in key:
            key = f"{key.rpartition('+')[0]}+upper_{key_no_modifiers}"
        else:
            key = f"upper_{key_no_modifiers}"
    return key.replace("+", "_").lower()


@rich.repr.auto
class MouseEvent(InputEvent, bubble=True):
    """Sent in response to a mouse event.

    - [X] Bubbles
    - [ ] Verbose

    Args:
        x: The relative x coordinate.
        y: The relative y coordinate.
        delta_x: Change in x since the last message.
        delta_y: Change in y since the last message.
        button: Indexed of the pressed button.
        shift: True if the shift key is pressed.
        meta: True if the meta key is pressed.
        ctrl: True if the ctrl key is pressed.
        screen_x: The absolute x coordinate.
        screen_y: The absolute y coordinate.
        style: The Rich Style under the mouse cursor.
    """

    __slots__ = [
        "x",
        "y",
        "delta_x",
        "delta_y",
        "button",
        "shift",
        "meta",
        "ctrl",
        "screen_x",
        "screen_y",
        "_style",
    ]

    def __init__(
        self,
        x: int,
        y: int,
        delta_x: int,
        delta_y: int,
        button: int,
        shift: bool,
        meta: bool,
        ctrl: bool,
        screen_x: int | None = None,
        screen_y: int | None = None,
        style: Style | None = None,
    ) -> None:
        super().__init__()
        self.x = x
        """The relative x coordinate."""
        self.y = y
        """The relative y coordinate."""
        self.delta_x = delta_x
        """Change in x since the last message."""
        self.delta_y = delta_y
        """Change in y since the last message."""
        self.button = button
        """Indexed of the pressed button."""
        self.shift = shift
        """`True` if the shift key is pressed."""
        self.meta = meta
        """`True` if the meta key is pressed."""
        self.ctrl = ctrl
        """`True` if the ctrl key is pressed."""
        self.screen_x = x if screen_x is None else screen_x
        """The absolute x coordinate."""
        self.screen_y = y if screen_y is None else screen_y
        """The absolute y coordinate."""
        self._style = style or Style()

    @classmethod
    def from_event(cls: Type[MouseEventT], event: MouseEvent) -> MouseEventT:
        new_event = cls(
            event.x,
            event.y,
            event.delta_x,
            event.delta_y,
            event.button,
            event.shift,
            event.meta,
            event.ctrl,
            event.screen_x,
            event.screen_y,
            event._style,
        )
        return new_event

    def __rich_repr__(self) -> rich.repr.Result:
        yield "x", self.x
        yield "y", self.y
        yield "delta_x", self.delta_x, 0
        yield "delta_y", self.delta_y, 0
        if self.screen_x != self.x:
            yield "screen_x", self.screen_x
        if self.screen_y != self.y:
            yield "screen_y", self.screen_y
        yield "button", self.button, 0
        yield "shift", self.shift, False
        yield "meta", self.meta, False
        yield "ctrl", self.ctrl, False

    @property
    def offset(self) -> Offset:
        """The mouse coordinate as an offset.

        Returns:
            Mouse coordinate.
        """
        return Offset(self.x, self.y)

    @property
    def screen_offset(self) -> Offset:
        """Mouse coordinate relative to the screen."""
        return Offset(self.screen_x, self.screen_y)

    @property
    def delta(self) -> Offset:
        """Mouse coordinate delta (change since last event)."""
        return Offset(self.delta_x, self.delta_y)

    @property
    def style(self) -> Style:
        """The (Rich) Style under the cursor."""
        return self._style or Style()

    @style.setter
    def style(self, style: Style) -> None:
        self._style = style

    def get_content_offset(self, widget: Widget) -> Offset | None:
        """Get offset within a widget's content area, or None if offset is not in content (i.e. padding or border).

        Args:
            widget: Widget receiving the event.

        Returns:
            An offset where the origin is at the top left of the content area.
        """
        if self.screen_offset not in widget.content_region:
            return None
        return self.get_content_offset_capture(widget)

    def get_content_offset_capture(self, widget: Widget) -> Offset:
        """Get offset from a widget's content area.

        This method works even if the offset is outside the widget content region.

        Args:
            widget: Widget receiving the event.

        Returns:
            An offset where the origin is at the top left of the content area.
        """
        return self.offset - widget.gutter.top_left

    def _apply_offset(self, x: int, y: int) -> MouseEvent:
        return self.__class__(
            x=self.x + x,
            y=self.y + y,
            delta_x=self.delta_x,
            delta_y=self.delta_y,
            button=self.button,
            shift=self.shift,
            meta=self.meta,
            ctrl=self.ctrl,
            screen_x=self.screen_x,
            screen_y=self.screen_y,
            style=self.style,
        )


@rich.repr.auto
class MouseMove(MouseEvent, bubble=True, verbose=True):
    """Sent when the mouse cursor moves.

    - [X] Bubbles
    - [X] Verbose
    """


@rich.repr.auto
class MouseDown(MouseEvent, bubble=True, verbose=True):
    """Sent when a mouse button is pressed.

    - [X] Bubbles
    - [X] Verbose
    """


@rich.repr.auto
class MouseUp(MouseEvent, bubble=True, verbose=True):
    """Sent when a mouse button is released.

    - [X] Bubbles
    - [X] Verbose
    """


@rich.repr.auto
class MouseScrollDown(MouseEvent, bubble=True, verbose=True):
    """Sent when the mouse wheel is scrolled *down*.

    - [X] Bubbles
    - [ ] Verbose
    """


@rich.repr.auto
class MouseScrollUp(MouseEvent, bubble=True, verbose=True):
    """Sent when the mouse wheel is scrolled *up*.

    - [X] Bubbles
    - [ ] Verbose
    """


class Click(MouseEvent, bubble=True):
    """Sent when a widget is clicked.

    - [X] Bubbles
    - [ ] Verbose
    """


@rich.repr.auto
class Timer(Event, bubble=False, verbose=True):
    """Sent in response to a timer.

    - [ ] Bubbles
    - [X] Verbose
    """

    __slots__ = ["time", "count", "callback"]

    def __init__(
        self,
        timer: "TimerClass",
        time: float,
        count: int = 0,
        callback: TimerCallback | None = None,
    ) -> None:
        super().__init__()
        self.timer = timer
        self.time = time
        self.count = count
        self.callback = callback

    def __rich_repr__(self) -> rich.repr.Result:
        yield self.timer.name
        yield "count", self.count


class Enter(Event, bubble=True, verbose=True):
    """Sent when the mouse is moved over a widget.

    Note that this event bubbles, so a widget may receive this event when the mouse
    moves over a child widget. Check the `node` attribute for the widget directly under
    the mouse.

    - [X] Bubbles
    - [X] Verbose
    """

    __slots__ = ["node"]

    def __init__(self, node: DOMNode) -> None:
        self.node = node
        """The node directly under the mouse."""
        super().__init__()


class Leave(Event, bubble=True, verbose=True):
    """Sent when the mouse is moved away from a widget, or if a widget is
    programmatically disabled while hovered.

    Note that this widget bubbles, so a widget may receive Leave events for any child widgets.
    Check the `node` parameter for the original widget that was previously under the mouse.


    - [X] Bubbles
    - [X] Verbose
    """

    __slots__ = ["node"]

    def __init__(self, node: DOMNode) -> None:
        self.node = node
        """The node that was previously directly under the mouse."""
        super().__init__()


class Focus(Event, bubble=False):
    """Sent when a widget is focussed.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Blur(Event, bubble=False):
    """Sent when a widget is blurred (un-focussed).

    - [ ] Bubbles
    - [ ] Verbose
    """


class AppFocus(Event, bubble=False):
    """Sent when the app has focus.

    - [ ] Bubbles
    - [ ] Verbose

    Note:
        Only available when running within a terminal that supports
        `FocusIn`, or when running via textual-web.
    """


class AppBlur(Event, bubble=False):
    """Sent when the app loses focus.

    - [ ] Bubbles
    - [ ] Verbose

    Note:
        Only available when running within a terminal that supports
        `FocusOut`, or when running via textual-web.
    """


@dataclass
class DescendantFocus(Event, bubble=True, verbose=True):
    """Sent when a child widget is focussed.

    - [X] Bubbles
    - [X] Verbose
    """

    widget: Widget
    """The widget that was focused."""

    @property
    def control(self) -> Widget:
        """The widget that was focused (alias of `widget`)."""
        return self.widget


@dataclass
class DescendantBlur(Event, bubble=True, verbose=True):
    """Sent when a child widget is blurred.

    - [X] Bubbles
    - [X] Verbose
    """

    widget: Widget
    """The widget that was blurred."""

    @property
    def control(self) -> Widget:
        """The widget that was blurred (alias of `widget`)."""
        return self.widget


@rich.repr.auto
class Paste(Event, bubble=True):
    """Event containing text that was pasted into the Textual application.
    This event will only appear when running in a terminal emulator that supports
    bracketed paste mode. Textual will enable bracketed pastes when an app starts,
    and disable it when the app shuts down.

    - [X] Bubbles
    - [ ] Verbose


    Args:
        text: The text that has been pasted.
    """

    def __init__(self, text: str) -> None:
        super().__init__()
        self.text = text
        """The text that was pasted."""

    def __rich_repr__(self) -> rich.repr.Result:
        yield "text", self.text


class ScreenResume(Event, bubble=False):
    """Sent to screen that has been made active.

    - [ ] Bubbles
    - [ ] Verbose
    """


class ScreenSuspend(Event, bubble=False):
    """Sent to screen when it is no longer active.

    - [ ] Bubbles
    - [ ] Verbose
    """


@rich.repr.auto
class Print(Event, bubble=False):
    """Sent to a widget that is capturing [`print`][print].

    - [ ] Bubbles
    - [ ] Verbose

    Args:
        text: Text that was printed.
        stderr: `True` if the print was to stderr, or `False` for stdout.

    Note:
        Python's [`print`][print] output can be captured with
        [`App.begin_capture_print`][textual.app.App.begin_capture_print].
    """

    def __init__(self, text: str, stderr: bool = False) -> None:
        super().__init__()
        self.text = text
        """The text that was printed."""
        self.stderr = stderr
        """`True` if the print was to stderr, or `False` for stdout."""

    def __rich_repr__(self) -> rich.repr.Result:
        yield self.text
        yield self.stderr


@dataclass
class DeliveryComplete(Event, bubble=False):
    """Sent to App when a file has been delivered."""

    key: str
    """The delivery key associated with the delivery.
    
    This is the same key that was returned by `App.deliver_text`/`App.deliver_binary`.
    """

    path: Path | None = None
    """The path where the file was saved, or `None` if the path is not available, for
    example if the file was delivered via web browser.
    """

    name: str | None = None
    """Optional name returned to the app to identify the download."""


@dataclass
class DeliveryFailed(Event, bubble=False):
    """Sent to App when a file delivery fails."""

    key: str
    """The delivery key associated with the delivery."""

    exception: BaseException
    """The exception that was raised during the delivery."""

    name: str | None = None
    """Optional name returned to the app to identify the download."""



import string

import pytest

from textual.app import App, ComposeResult
from textual.suggester import SuggestFromList
from textual.widgets import Input


class SuggestionsApp(App[ComposeResult]):
    def __init__(self, suggestions):
        self.suggestions = suggestions
        self.input = Input(suggester=SuggestFromList(self.suggestions))
        super().__init__()

    def compose(self) -> ComposeResult:
        yield self.input


async def test_no_suggestions():
    app = SuggestionsApp([])
    async with app.run_test() as pilot:
        assert app.input._suggestion == ""
        await pilot.press("a")
        assert app.input._suggestion == ""


async def test_suggestion():
    app = SuggestionsApp(["hello"])
    async with app.run_test() as pilot:
        for char in "hello":
            await pilot.press(char)
            assert app.input._suggestion == "hello"


async def test_accept_suggestion():
    app = SuggestionsApp(["hello"])
    async with app.run_test() as pilot:
        await pilot.press("h")
        await pilot.press("right")
        assert app.input.value == "hello"


async def test_no_suggestion_on_empty_value():
    app = SuggestionsApp(["hello"])
    async with app.run_test():
        assert app.input._suggestion == ""


async def test_no_suggestion_on_empty_value_after_deleting():
    app = SuggestionsApp(["hello"])
    async with app.run_test() as pilot:
        await pilot.press("h", "e", "backspace", "backspace")
        assert app.input.value == ""  # Sanity check.
        assert app.input._suggestion == ""


async def test_suggestion_shows_up_after_deleting_extra_chars():
    app = SuggestionsApp(["hello"])
    async with app.run_test() as pilot:
        await pilot.press(*"help")
        assert app.input._suggestion == ""
        await pilot.press("backspace")
        assert app.input._suggestion == "hello"


async def test_suggestion_shows_up_after_deleting_extra_chars_in_middle_of_word():
    app = SuggestionsApp(["hello"])
    async with app.run_test() as pilot:
        await pilot.press(*"hefl")
        assert app.input._suggestion == ""
        await pilot.press("left", "backspace")
        assert app.input._suggestion == "hello"


@pytest.mark.parametrize(
    ("suggestion", "truncate_at"),
    [
        (".......", 3),
        ("hey there", 3),
        ("Olá, tudo bem?", 3),
        ("áàóãõñç", 2),
        (string.punctuation, 3),
        (string.punctuation[::-1], 5),
        (string.punctuation[::3], 5),
    ],
)
async def test_suggestion_with_special_characters(suggestion: str, truncate_at: int):
    app = SuggestionsApp([suggestion])
    async with app.run_test() as pilot:
        await pilot.press(*suggestion[:truncate_at])
        assert app.input._suggestion == suggestion


async def test_suggestion_priority():
    app = SuggestionsApp(["dog", "dad"])
    async with app.run_test() as pilot:
        await pilot.press("d")
        assert app.input._suggestion == "dog"
        await pilot.press("a")
        assert app.input._suggestion == "dad"



from __future__ import annotations

from textual.app import App, ComposeResult
from textual.widgets import RadioButton, RadioSet


class RadioSetApp(App[None]):
    def __init__(self):
        super().__init__()
        self.events_received = []

    def compose(self) -> ComposeResult:
        with RadioSet(id="from_buttons"):
            yield RadioButton(id="clickme")
            yield RadioButton()
            yield RadioButton(value=True)
        yield RadioSet("One", "True", "Three", id="from_strings")

    def on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        assert event.radio_set is event.control
        self.events_received.append(
            (
                event.radio_set.id,
                event.index,
                [button.value for button in event.radio_set.query(RadioButton)],
            )
        )


async def test_radio_sets_initial_state():
    """The initial states of the radio sets should be as we specified."""
    async with RadioSetApp().run_test() as pilot:
        assert pilot.app.query_one("#from_buttons", RadioSet).pressed_index == 2
        assert pilot.app.query_one("#from_buttons", RadioSet).pressed_button is not None
        assert pilot.app.query_one("#from_strings", RadioSet).pressed_index == -1
        assert pilot.app.query_one("#from_strings", RadioSet).pressed_button is None
        assert pilot.app.events_received == []


async def test_click_sets_focus():
    """Clicking within a radio set should set focus."""
    async with RadioSetApp().run_test() as pilot:
        pilot.app.set_focus(None)
        assert pilot.app.screen.focused is None
        await pilot.click("#clickme")
        assert pilot.app.screen.focused == pilot.app.query_one("#from_buttons")


async def test_radio_sets_toggle():
    """Test the status of the radio sets after they've been toggled."""
    async with RadioSetApp().run_test() as pilot:
        pilot.app.query_one("#from_buttons", RadioSet)._nodes[0].toggle()
        pilot.app.query_one("#from_strings", RadioSet)._nodes[2].toggle()
        await pilot.pause()
        assert pilot.app.query_one("#from_buttons", RadioSet).pressed_index == 0
        assert pilot.app.query_one("#from_buttons", RadioSet).pressed_button is not None
        assert pilot.app.query_one("#from_strings", RadioSet).pressed_index == 2
        assert pilot.app.query_one("#from_strings", RadioSet).pressed_button is not None
        assert pilot.app.events_received == [
            ("from_buttons", 0, [True, False, False]),
            ("from_strings", 2, [False, False, True]),
        ]


async def test_radioset_same_button_mash():
    """Mashing the same button should have no effect."""
    async with RadioSetApp().run_test() as pilot:
        assert pilot.app.query_one("#from_buttons", RadioSet).pressed_index == 2
        pilot.app.query_one("#from_buttons", RadioSet)._nodes[2].toggle()
        assert pilot.app.query_one("#from_buttons", RadioSet).pressed_index == 2
        assert pilot.app.events_received == []


async def test_radioset_inner_navigation():
    """Using the cursor keys should navigate between buttons in a set."""
    async with RadioSetApp().run_test() as pilot:
        for key, landing in (
            ("down", 1),
            ("up", 0),
            ("right", 1),
            ("left", 0),
            ("up", 2),
            ("down", 0),
        ):
            await pilot.press(key, "enter")
            assert (
                pilot.app.query_one("#from_buttons", RadioSet).pressed_button
                == pilot.app.query_one("#from_buttons").children[landing]
            )
    async with RadioSetApp().run_test() as pilot:
        assert pilot.app.screen.focused is pilot.app.screen.query_one("#from_buttons")
        await pilot.press("tab")
        assert pilot.app.screen.focused is pilot.app.screen.query_one("#from_strings")
        assert pilot.app.query_one("#from_strings", RadioSet)._selected == 0
        await pilot.press("down")
        assert pilot.app.query_one("#from_strings", RadioSet)._selected == 1


async def test_radioset_inner_navigation_post_build():
    class EmptyRadioSetApp(App[None]):
        def compose(self) -> ComposeResult:
            yield RadioSet()

        def on_mount(self) -> None:
            # This isn't encouraged; but neither is it currently prohibited;
            # so let's test.
            for n in range(5):
                self.query_one(RadioSet).mount(RadioButton(id=f"rb{n}"))

    async with EmptyRadioSetApp().run_test() as pilot:
        assert pilot.app.query_one(RadioSet)._selected is None
        await pilot.press("up")
        assert pilot.app.query_one(RadioSet)._selected == 4


async def test_radioset_breakout_navigation():
    """Shift/Tabbing while in a radioset should move to the previous/next focsuable after the set itself."""
    async with RadioSetApp().run_test() as pilot:
        assert pilot.app.screen.focused is pilot.app.query_one("#from_buttons")
        await pilot.press("tab")
        assert pilot.app.screen.focused is pilot.app.query_one("#from_strings")
        await pilot.press("shift+tab")
        assert pilot.app.screen.focused is pilot.app.query_one("#from_buttons")


class BadRadioSetApp(App[None]):
    def compose(self) -> ComposeResult:
        with RadioSet():
            for n in range(20):
                yield RadioButton(str(n), True)


async def test_there_can_only_be_one():
    """Adding multiple 'on' buttons should result in only one on."""
    async with BadRadioSetApp().run_test() as pilot:
        assert len(pilot.app.query("RadioButton.-on")) == 1
        assert pilot.app.query_one(RadioSet).pressed_index == 0


class RadioSetDisabledButtonsApp(App[None]):
    def compose(self) -> ComposeResult:
        self.selected = []
        with RadioSet():
            yield RadioButton("0", disabled=True)
            yield RadioButton("1")
            yield RadioButton("2", disabled=True)
            yield RadioButton("3", disabled=True)
            yield RadioButton("4")
            yield RadioButton("5")
            yield RadioButton("6", disabled=True)
            yield RadioButton("7")
            yield RadioButton("8", disabled=True)

    def on_radio_set_changed(self, radio_set: RadioSet.Changed) -> None:
        self.selected.append(str(radio_set.pressed.label))


async def test_keyboard_navigation_with_disabled_buttons():
    """Regression test for https://github.com/Textualize/textual/issues/3839."""

    app = RadioSetDisabledButtonsApp()
    async with app.run_test() as pilot:
        await pilot.press("enter")
        for _ in range(5):
            await pilot.press("down")
            await pilot.press("enter")
        for _ in range(5):
            await pilot.press("up")
            await pilot.press("enter")

    assert app.selected == [
        "1",
        "4",
        "5",
        "7",
        "1",
        "4",
        "1",
        "7",
        "5",
        "4",
        "1",
    ]



from textual.app import App, ComposeResult
from textual.geometry import Offset
from textual.widgets import Input


class InputApp(App):
    # Apply padding to ensure gutter accounted for.
    CSS = "Input { padding: 4 8 }"

    def compose(self) -> ComposeResult:
        yield Input("こんにちは!")


async def test_initial_terminal_cursor_position():
    app = InputApp()
    async with app.run_test():
        # The input is focused so the terminal cursor position should update.
        assert app.cursor_position == Offset(21, 5)


async def test_terminal_cursor_position_update_on_cursor_move():
    app = InputApp()
    async with app.run_test():
        input_widget = app.query_one(Input)
        input_widget.action_cursor_left()
        input_widget.action_cursor_left()
        # We went left over two double-width characters
        assert app.cursor_position == Offset(17, 5)



from __future__ import annotations

from typing import Tuple

from typing_extensions import Literal

from textual.color import Color

DockEdge = Literal["none", "top", "right", "bottom", "left"]
EdgeType = Literal[
    "",
    "ascii",
    "none",
    "hidden",
    "blank",
    "round",
    "solid",
    "thick",
    "double",
    "dashed",
    "heavy",
    "inner",
    "outer",
    "hkey",
    "vkey",
    "tall",
    "panel",
    "wide",
]
Visibility = Literal["visible", "hidden", "initial", "inherit"]
Display = Literal["block", "none"]
AlignHorizontal = Literal["left", "center", "right"]
AlignVertical = Literal["top", "middle", "bottom"]
ScrollbarGutter = Literal["auto", "stable"]
BoxSizing = Literal["border-box", "content-box"]
Overflow = Literal["scroll", "hidden", "auto"]
EdgeStyle = Tuple[EdgeType, Color]
TextAlign = Literal["left", "start", "center", "right", "end", "justify"]
Constrain = Literal["none", "x", "y", "both"]
Overlay = Literal["none", "screen"]

Specificity3 = Tuple[int, int, int]
Specificity6 = Tuple[int, int, int, int, int, int]

CSSLocation = Tuple[str, str]
"""Represents the definition location of a piece of CSS code.

The first element of the tuple is the file path from where the CSS was read.
If the CSS was read from a Python source file, the second element contains the class
variable from where the CSS was read (e.g., "Widget.DEFAULT_CSS"), otherwise it's an
empty string.
"""



from textual.app import App, ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Button, TextArea


class TextAreaDialog(ModalScreen):
    BINDINGS = [("escape", "dismiss")]

    def compose(self) -> ComposeResult:
        yield TextArea(
            tab_behavior="focus",  # the default
        )
        yield Button("Submit")


class TextAreaDialogApp(App):
    def on_mount(self) -> None:
        self.push_screen(TextAreaDialog())


async def test_escape_key_when_tab_behavior_is_focus():
    """Regression test for https://github.com/Textualize/textual/issues/4110

    When the `tab_behavior` of TextArea is the default to shift focus,
    pressing <Escape> should not shift focus but instead skip and allow any
    parent bindings to run.
    """

    app = TextAreaDialogApp()
    async with app.run_test() as pilot:
        # Sanity check
        assert isinstance(pilot.app.screen, TextAreaDialog)
        assert isinstance(pilot.app.focused, TextArea)

        # Pressing escape should dismiss the dialog screen, not focus the button
        await pilot.press("escape")
        assert not isinstance(pilot.app.screen, TextAreaDialog)


async def test_escape_key_when_tab_behavior_is_indent():
    """When the `tab_behavior` of TextArea is indent rather than switch focus,
    pressing <Escape> should instead shift focus.
    """

    app = TextAreaDialogApp()
    async with app.run_test() as pilot:
        # Sanity check
        assert isinstance(pilot.app.screen, TextAreaDialog)
        assert isinstance(pilot.app.focused, TextArea)

        pilot.app.query_one(TextArea).tab_behavior = "indent"
        # Pressing escape should focus the button, not dismiss the dialog screen
        await pilot.press("escape")
        assert isinstance(pilot.app.screen, TextAreaDialog)
        assert isinstance(pilot.app.focused, Button)



from textual.app import App, ComposeResult
from textual.widgets import Header, Select

LINES = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
I will permit it to pass over me and through me.""".splitlines()


async def test_select_remove():
    # Regression test for https://github.com/Textualize/textual/issues/4782
    class SelectApp(App):
        def compose(self) -> ComposeResult:
            self.select = Select((line, line) for line in LINES)
            self.select.watch_value = self.on_select
            yield Header()
            yield self.select

        def on_select(self):
            self.select.remove()

    app = SelectApp()
    async with app.run_test() as pilot:
        await pilot.press("enter", "down", "down", "enter")



import pytest

from textual.app import App, ComposeResult
from textual.color import Color
from textual.containers import Container
from textual.css.query import (
    DeclarationError,
    InvalidQueryFormat,
    NoMatches,
    TooManyMatches,
    WrongType,
)
from textual.widget import Widget
from textual.widgets import Input, Label


def test_query():
    class View(Widget):
        pass

    class View2(View):
        pass

    class App(Widget):
        pass

    app = App()
    main_view = View(id="main")
    help_view = View2(id="help")
    app._add_child(main_view)
    app._add_child(help_view)

    widget1 = Widget(id="widget1")
    widget2 = Widget(id="widget2")
    sidebar = Widget(id="sidebar")
    sidebar.add_class("float")

    helpbar = Widget(id="helpbar")
    helpbar.add_class("float")

    main_view._add_child(widget1)
    main_view._add_child(widget2)
    main_view._add_child(sidebar)

    sub_view = View(id="sub")
    sub_view.add_class("-subview")
    main_view._add_child(sub_view)

    tooltip = Widget(id="tooltip")
    tooltip.add_class("float", "transient")
    sub_view._add_child(tooltip)

    help = Widget(id="markdown")
    help_view._add_child(help)
    help_view._add_child(helpbar)

    # repeat tests to account for caching
    for repeat in range(3):
        assert list(app.query_children()) == [main_view, help_view]
        assert list(app.query_children("*")) == [main_view, help_view]
        assert list(app.query_children("#help")) == [help_view]
        assert list(main_view.query_children(".float")) == [sidebar]

        assert list(app.query("Frob")) == []
        assert list(app.query(".frob")) == []
        assert list(app.query("#frob")) == []

        assert not app.query("NotAnApp")

        assert list(app.query("App")) == []  # Root is not included in queries
        assert list(app.query("#main")) == [main_view]
        assert list(app.query("View#main")) == [main_view]
        assert list(app.query("View2#help")) == [help_view]
        assert list(app.query("#widget1")) == [widget1]
        assert list(app.query("#Widget1")) == []  # Note case.
        assert list(app.query("#widget2")) == [widget2]
        assert list(app.query("#Widget2")) == []  # Note case.

        assert list(app.query("Widget.float")) == [sidebar, tooltip, helpbar]
        assert list(app.query(Widget).filter(".float")) == [sidebar, tooltip, helpbar]
        assert list(
            app.query(Widget)
            .exclude("App")
            .exclude("#sub")
            .exclude("#markdown")
            .exclude("#main")
            .exclude("#help")
            .exclude("#widget1")
            .exclude("#widget2")
        ) == [sidebar, tooltip, helpbar]
        assert list(reversed(app.query("Widget.float"))) == [helpbar, tooltip, sidebar]
        assert list(app.query("Widget.float").results(Widget)) == [
            sidebar,
            tooltip,
            helpbar,
        ]
        assert list(app.query("Widget.float").results()) == [
            sidebar,
            tooltip,
            helpbar,
        ]
        assert list(app.query("Widget.float").results(View)) == []
        assert app.query_one("#widget1") == widget1
        assert app.query_one("#widget1", Widget) == widget1
        with pytest.raises(TooManyMatches):
            _ = app.query_exactly_one(Widget)

        assert app.query("Widget.float")[0] == sidebar
        assert app.query("Widget.float")[0:2] == [sidebar, tooltip]

        assert list(app.query("Widget.float.transient")) == [tooltip]

        assert list(app.query("App > View")) == [main_view, help_view]
        assert list(app.query("App > View#help")) == [help_view]
        assert list(app.query("App > View#main .float ")) == [sidebar, tooltip]
        assert list(app.query("View > View")) == [sub_view]

        assert list(app.query("#help *")) == [help, helpbar]
        assert list(app.query("#main *")) == [
            widget1,
            widget2,
            sidebar,
            sub_view,
            tooltip,
        ]

        assert list(app.query("View")) == [main_view, sub_view, help_view]
        assert list(app.query("#widget1, #widget2")) == [widget1, widget2]
        assert list(app.query("#widget1 , #widget2")) == [widget1, widget2]
        assert list(app.query("#widget1, #widget2, App")) == [widget1, widget2]

        assert app.query(".float").first() == sidebar
        assert app.query(".float").last() == helpbar

        with pytest.raises(NoMatches):
            _ = app.query(".no_such_class").first()
        with pytest.raises(NoMatches):
            _ = app.query(".no_such_class").last()

        with pytest.raises(WrongType):
            _ = app.query(".float").first(View)
        with pytest.raises(WrongType):
            _ = app.query(".float").last(View)


def test_query_classes():
    class App(Widget):
        pass

    class ClassTest(Widget):
        pass

    CHILD_COUNT = 100

    # Create a fake app to hold everything else.
    app = App()

    # Now spin up a bunch of children.
    for n in range(CHILD_COUNT):
        app._add_child(ClassTest(id=f"child{n}"))

    # Let's just be 100% sure everything was created fine.
    assert len(app.query(ClassTest)) == CHILD_COUNT

    # Now, let's check there are *no* children with the test class.
    assert len(app.query(".test")) == 0

    # Add classes via set_classes
    app.query(ClassTest).set_classes("foo bar")
    assert (len(app.query(".foo"))) == CHILD_COUNT
    assert (len(app.query(".bar"))) == CHILD_COUNT

    # Reset classes
    app.query(ClassTest).set_classes("")
    assert (len(app.query(".foo"))) == 0
    assert (len(app.query(".bar"))) == 0

    # Repeat, to check setting empty iterable
    app.query(ClassTest).set_classes("foo bar")
    assert (len(app.query(".foo"))) == CHILD_COUNT
    assert (len(app.query(".bar"))) == CHILD_COUNT

    app.query(ClassTest).set_classes([])
    assert (len(app.query(".foo"))) == 0
    assert (len(app.query(".bar"))) == 0

    # Add the test class to everything and then check again.
    app.query(ClassTest).add_class("test")
    assert len(app.query(".test")) == CHILD_COUNT

    # Remove the test class from everything then try again.
    app.query(ClassTest).remove_class("test")
    assert len(app.query(".test")) == 0

    # Add the test class to everything using set_class.
    app.query(ClassTest).set_class(True, "test")
    assert len(app.query(".test")) == CHILD_COUNT

    # Remove the test class from everything using set_class.
    app.query(ClassTest).set_class(False, "test")
    assert len(app.query(".test")) == 0

    # Add the test class to everything using toggle_class.
    app.query(ClassTest).toggle_class("test")
    assert len(app.query(".test")) == CHILD_COUNT

    # Remove the test class from everything using toggle_class.
    app.query(ClassTest).toggle_class("test")
    assert len(app.query(".test")) == 0


def test_invalid_query():
    class App(Widget):
        pass

    app = App()

    with pytest.raises(InvalidQueryFormat):
        app.query("#3")

    with pytest.raises(InvalidQueryFormat):
        app.query("#foo").exclude("#2")


async def test_universal_selector_doesnt_select_self():
    class ExampleApp(App):
        def compose(self) -> ComposeResult:
            yield Container(
                Widget(
                    Widget(),
                    Widget(
                        Widget(),
                    ),
                ),
                Widget(),
                id="root-container",
            )

    app = ExampleApp()
    async with app.run_test():
        container = app.query_one("#root-container", Container)
        query = container.query("*")
        results = list(query.results())
        assert len(results) == 5
        assert not any(node.id == "root-container" for node in results)


async def test_query_set_styles_invalid_css_raises_error():
    app = App()
    async with app.run_test():
        with pytest.raises(DeclarationError):
            app.query(Widget).set_styles(css="random-rule: 1fr;")


async def test_query_set_styles_kwds():
    class LabelApp(App):
        def compose(self):
            yield Label("Some text")

    app = LabelApp()
    async with app.run_test():
        # Sanity check.
        assert app.query_one(Label).styles.color != Color(255, 0, 0)
        app.query(Label).set_styles(color="red")
        assert app.query_one(Label).styles.color == Color(255, 0, 0)


async def test_query_set_styles_css_and_kwds():
    class LabelApp(App):
        def compose(self):
            yield Label("Some text")

    app = LabelApp()
    async with app.run_test():
        # Sanity checks.
        lbl = app.query_one(Label)
        assert lbl.styles.color != Color(255, 0, 0)
        assert lbl.styles.background != Color(255, 0, 0)

        app.query(Label).set_styles(css="background: red;", color="red")
        assert app.query_one(Label).styles.color == Color(255, 0, 0)
        assert app.query_one(Label).styles.background == Color(255, 0, 0)


async def test_query_set_styles_css():
    class LabelApp(App):
        def compose(self):
            yield Label("Some text")

    app = LabelApp()
    async with app.run_test():
        # Sanity checks.
        lbl = app.query_one(Label)
        assert lbl.styles.color != Color(255, 0, 0)
        assert lbl.styles.background != Color(255, 0, 0)

        app.query(Label).set_styles(css="background: red; color: red;")
        assert app.query_one(Label).styles.color == Color(255, 0, 0)
        assert app.query_one(Label).styles.background == Color(255, 0, 0)


@pytest.mark.parametrize(
    "args", [(False, False), (True, False), (True, True), (False, True)]
)
async def test_query_refresh(args):
    refreshes = []

    class MyWidget(Widget):
        def refresh(self, *, repaint=None, layout=None, recompose=None):
            super().refresh(repaint=repaint, layout=layout)
            refreshes.append((repaint, layout))

    class MyApp(App):
        def compose(self):
            yield MyWidget()

    app = MyApp()
    async with app.run_test() as pilot:
        app.query(MyWidget).refresh(repaint=args[0], layout=args[1])
        assert refreshes[-1] == args


async def test_query_focus_blur():
    class FocusApp(App):
        AUTO_FOCUS = None

        def compose(self) -> ComposeResult:
            yield Input(id="foo")
            yield Input(id="bar")
            yield Input(id="baz")

    app = FocusApp()
    async with app.run_test() as pilot:
        # Nothing focused
        assert app.focused is None
        # Focus first input
        app.query(Input).focus()
        await pilot.pause()
        assert app.focused.id == "foo"
        # Blur inputs
        app.query(Input).blur()
        await pilot.pause()
        assert app.focused is None
        # Focus another
        app.query("#bar").focus()
        await pilot.pause()
        assert app.focused.id == "bar"
        # Focus non existing
        app.query("#egg").focus()
        assert app.focused.id == "bar"



from textual.app import App, ComposeResult
from textual.widgets import ProgressBar


class TooltipApp(App[None]):
    TOOLTIP_DELAY = 0.4
    def compose(self) -> ComposeResult:
        progress_bar = ProgressBar(100, show_eta=False)
        progress_bar.advance(10)
        progress_bar.tooltip = "Hello, Tooltip!"
        yield progress_bar


if __name__ == "__main__":
    app = TooltipApp()
    app.run()



from itertools import cycle

from textual.app import App, ComposeResult
from textual.widgets import Static

hellos = cycle(
    [
        "Hola",
        "Bonjour",
        "Guten tag",
        "Salve",
        "Nǐn hǎo",
        "Olá",
        "Asalaam alaikum",
        "Konnichiwa",
        "Anyoung haseyo",
        "Zdravstvuyte",
        "Hello",
    ]
)


class Hello(Static):
    """Display a greeting."""

    def on_mount(self) -> None:
        self.action_next_word()

    def action_next_word(self) -> None:
        """Get a new hello and update the content area."""
        hello = next(hellos)
        self.update(f"[@click='next_word']{hello}[/], [b]World[/b]!")


class CustomApp(App):
    CSS_PATH = "hello05.tcss"

    def compose(self) -> ComposeResult:
        yield Hello()


if __name__ == "__main__":
    app = CustomApp()
    app.run()



from __future__ import annotations

from bisect import bisect_right

from rich.text import Text

from textual._cells import cell_len, cell_width_to_column_index
from textual._wrap import compute_wrap_offsets
from textual.document._document import DocumentBase, Location
from textual.expand_tabs import expand_tabs_inline, get_tab_widths
from textual.geometry import Offset, clamp

VerticalOffset = int
LineIndex = int
SectionOffset = int


class WrappedDocument:
    """A view into a Document which wraps the document at a certain
    width and can be queried to retrieve lines from the *wrapped* version
    of the document.

    Allows for incremental updates, ensuring that we only re-wrap ranges of the document
    that were influenced by edits.
    """

    def __init__(
        self,
        document: DocumentBase,
        width: int = 0,
        tab_width: int = 4,
    ) -> None:
        """Construct a WrappedDocument.

        By default, a WrappedDocument is wrapped with width=0 (no wrapping).
        To wrap the document, use the wrap() method.

        Args:
            document: The document to wrap.
            width: The width to wrap at.
            tab_width: The maximum width to consider for tab characters.
        """
        self.document = document
        """The document wrapping is performed on."""

        self._wrap_offsets: list[list[int]] = []
        """Maps line indices to the offsets within the line where wrapping
        breaks should be added."""

        self._tab_width_cache: list[list[int]] = []
        """Maps line indices to a list of tab widths. `[[2, 4]]` means that on line 0, the first
        tab has width 2, and the second tab has width 4."""

        self._offset_to_line_info: list[tuple[LineIndex, SectionOffset]] = []
        """Maps y_offsets (from the top of the document) to line_index and the offset
        of the section within the line."""

        self._line_index_to_offsets: list[list[VerticalOffset]] = []
        """Maps line indices to all the vertical offsets which correspond to that line."""

        self._width: int = width
        """The width the document is currently wrapped at. This will correspond with
        the value last passed into the `wrap` method."""

        self._tab_width: int = tab_width
        """The maximum width to expand tabs to when considering their widths."""

        self.wrap(width, tab_width)

    @property
    def wrapped(self) -> bool:
        """True if the content is wrapped. This is not the same as wrapping being "enabled".
        For example, an empty document can have wrapping enabled, but no wrapping has actually
        occurred.

        In other words, this is True if the length of any line in the document is greater
        than the available width."""
        return len(self._line_index_to_offsets) == len(self._offset_to_line_info)

    def wrap(self, width: int, tab_width: int | None = None) -> None:
        """Wrap and cache all lines in the document.

        Args:
            width: The width to wrap at. 0 for no wrapping.
            tab_width: The maximum width to consider for tab characters. If None,
                reuse the  tab width.
        """
        self._width = width
        if tab_width:
            self._tab_width = tab_width

        # We're starting wrapping from scratch
        new_wrap_offsets: list[list[int]] = []
        offset_to_line_info: list[tuple[LineIndex, SectionOffset]] = []
        line_index_to_offsets: list[list[VerticalOffset]] = []
        line_tab_widths: list[list[int]] = []

        append_wrap_offset = new_wrap_offsets.append
        append_line_info = offset_to_line_info.append
        append_line_offsets = line_index_to_offsets.append
        append_line_tab_widths = line_tab_widths.append

        current_offset = 0
        tab_width = self._tab_width
        for line_index, line in enumerate(self.document.lines):
            tab_sections = get_tab_widths(line, tab_width)
            wrap_offsets = (
                compute_wrap_offsets(
                    line,
                    width,
                    tab_size=tab_width,
                    precomputed_tab_sections=tab_sections,
                )
                if width
                else []
            )
            append_line_tab_widths([width for _, width in tab_sections])
            append_wrap_offset(wrap_offsets)
            append_line_offsets([])
            for section_y_offset in range(len(wrap_offsets) + 1):
                append_line_info((line_index, section_y_offset))
                line_index_to_offsets[line_index].append(current_offset)
                current_offset += 1

        self._wrap_offsets = new_wrap_offsets
        self._offset_to_line_info = offset_to_line_info
        self._line_index_to_offsets = line_index_to_offsets
        self._tab_width_cache = line_tab_widths

    @property
    def lines(self) -> list[list[str]]:
        """The lines of the wrapped version of the Document.

        Each index in the returned list represents a line index in the raw
        document. The list[str] at each index is the content of the raw document line
        split into multiple lines via wrapping.

        Note that this is expensive to compute and is not cached.

        Returns:
            A list of lines from the wrapped version of the document.
        """
        wrapped_lines: list[list[str]] = []
        append = wrapped_lines.append
        for line_index, line in enumerate(self.document.lines):
            divided = Text(line).divide(self._wrap_offsets[line_index])
            append([section.plain for section in divided])

        return wrapped_lines

    @property
    def height(self) -> int:
        """The height of the wrapped document."""
        return sum(len(offsets) + 1 for offsets in self._wrap_offsets)

    def wrap_range(
        self,
        start: Location,
        old_end: Location,
        new_end: Location,
    ) -> None:
        """Incrementally recompute wrapping based on a performed edit.

        This must be called *after* the source document has been edited.

        Args:
            start: The start location of the edit that was performed in document-space.
            old_end: The old end location of the edit in document-space.
            new_end: The new end location of the edit in document-space.
        """
        start_line_index, _ = start
        old_end_line_index, _ = old_end
        new_end_line_index, _ = new_end

        # Although end users should not be able to edit invalid ranges via a TextArea,
        #  programmers can pass whatever they wish to the edit API, so we need to clamp
        #  the edit ranges here to ensure we only attempt to update within the bounds
        #  of the wrapped document.
        old_max_index = len(self._line_index_to_offsets) - 1
        new_max_index = self.document.line_count - 1

        start_line_index = clamp(
            start_line_index, 0, min((old_max_index, new_max_index))
        )
        old_end_line_index = clamp(old_end_line_index, 0, old_max_index)
        new_end_line_index = clamp(new_end_line_index, 0, new_max_index)

        top_line_index, old_bottom_line_index = sorted(
            (start_line_index, old_end_line_index)
        )
        new_bottom_line_index = max((start_line_index, new_end_line_index))

        top_y_offset = self._line_index_to_offsets[top_line_index][0]
        old_bottom_y_offset = self._line_index_to_offsets[old_bottom_line_index][-1]

        # Get the new range of the edit from top to bottom.
        new_lines = self.document.lines[top_line_index : new_bottom_line_index + 1]

        new_wrap_offsets: list[list[int]] = []
        new_line_index_to_offsets: list[list[VerticalOffset]] = []
        new_offset_to_line_info: list[tuple[LineIndex, SectionOffset]] = []
        new_tab_widths: list[list[int]] = []

        append_wrap_offsets = new_wrap_offsets.append
        append_tab_widths = new_tab_widths.append

        width = self._width
        tab_width = self._tab_width

        # Add the new offsets between the top and new bottom (the new post-edit offsets)
        current_y_offset = top_y_offset
        for line_index, line in enumerate(new_lines, top_line_index):
            tab_sections = get_tab_widths(line, tab_width)
            wrap_offsets = (
                compute_wrap_offsets(
                    line, width, tab_width, precomputed_tab_sections=tab_sections
                )
                if width
                else []
            )
            append_tab_widths([width for _, width in tab_sections])
            append_wrap_offsets(wrap_offsets)

            # Collect up the new y offsets for this document line
            y_offsets_for_line: list[int] = []
            for section_offset in range(len(wrap_offsets) + 1):
                y_offsets_for_line.append(current_y_offset)
                new_offset_to_line_info.append((line_index, section_offset))
                current_y_offset += 1

            # Save the new y offsets for this line
            new_line_index_to_offsets.append(y_offsets_for_line)

        # Replace the range start -> old with the new wrapped lines
        self._offset_to_line_info[top_y_offset : old_bottom_y_offset + 1] = (
            new_offset_to_line_info
        )

        self._line_index_to_offsets[top_line_index : old_bottom_line_index + 1] = (
            new_line_index_to_offsets
        )

        self._tab_width_cache[top_line_index : old_bottom_line_index + 1] = (
            new_tab_widths
        )

        # How much did the edit/rewrap alter the offsets?
        old_height = old_bottom_y_offset - top_y_offset + 1
        new_height = len(new_offset_to_line_info)

        offset_shift = new_height - old_height
        line_shift = new_bottom_line_index - old_bottom_line_index

        # Update the line info at all offsets below the edit region.
        if line_shift:
            for y_offset in range(
                top_y_offset + new_height, len(self._offset_to_line_info)
            ):
                old_line_index, section_offset = self._offset_to_line_info[y_offset]
                new_line_index = old_line_index + line_shift
                new_line_info = (new_line_index, section_offset)
                self._offset_to_line_info[y_offset] = new_line_info

        # Update the offsets at all lines below the edit region
        if offset_shift:
            for line_index in range(
                top_line_index + len(new_lines), len(self._line_index_to_offsets)
            ):
                old_offsets = self._line_index_to_offsets[line_index]
                new_offsets = [offset + offset_shift for offset in old_offsets]
                self._line_index_to_offsets[line_index] = new_offsets

        self._wrap_offsets[top_line_index : old_bottom_line_index + 1] = (
            new_wrap_offsets
        )

    def offset_to_location(self, offset: Offset) -> Location:
        """Given an offset within the wrapped/visual display of the document,
        return the corresponding location in the document.

        Args:
            offset: The y-offset within the document.

        Raises:
            ValueError: When the given offset does not correspond to a line
                in the document.

        Returns:
            The Location in the document corresponding to the given offset.
        """
        x, y = offset
        x = max(0, x)
        y = max(0, y)

        if not self._width:
            # No wrapping, so we directly map offset to location and clamp.
            line_index = min(y, len(self._wrap_offsets) - 1)
            column_index = min(x, len(self.document.get_line(line_index)))
            return line_index, column_index

        # Find the line corresponding to the given y offset in the wrapped document.
        get_target_document_column = self.get_target_document_column

        try:
            offset_data = self._offset_to_line_info[y]
        except IndexError:
            # y-offset is too large
            offset_data = self._offset_to_line_info[-1]

        if offset_data is not None:
            line_index, section_y = offset_data
            location = line_index, get_target_document_column(
                line_index,
                x,
                section_y,
            )
        else:
            location = len(self._wrap_offsets) - 1, get_target_document_column(
                -1, x, -1
            )

        # Offset doesn't match any line => land on bottom wrapped line
        return location

    def location_to_offset(self, location: Location) -> Offset:
        """
        Convert a location in the document to an offset within the wrapped/visual display of the document.

        Args:
            location: The location in the document.

        Returns:
            The Offset in the document's visual display corresponding to the given location.
        """
        line_index, column_index = location

        # Clamp the line index to the bounds of the document
        line_index = clamp(line_index, 0, len(self._line_index_to_offsets))

        # Find the section index of this location, so that we know which y_offset to use
        wrap_offsets = self.get_offsets(line_index)
        section_start_columns = [0, *wrap_offsets]
        section_index = bisect_right(wrap_offsets, column_index)

        # Get the y-offsets corresponding to this line index
        y_offsets = self._line_index_to_offsets[line_index]
        section_column_index = column_index - section_start_columns[section_index]

        section = self.get_sections(line_index)[section_index]
        x_offset = cell_len(
            expand_tabs_inline(section[:section_column_index], self._tab_width)
        )

        return Offset(x_offset, y_offsets[section_index])

    def get_target_document_column(
        self,
        line_index: int,
        x_offset: int,
        y_offset: int,
    ) -> int:
        """Given a line index and the offsets within the wrapped version of that
        line, return the corresponding column index in the raw document.

        Args:
             line_index: The index of the line in the document.
             x_offset: The x-offset within the wrapped line.
             y_offset: The y-offset within the wrapped line (supports negative indexing).

        Returns:
            The column index corresponding to the line index and y offset.
        """

        # We've found the relevant line, now find the character by
        # looking at the character corresponding to the offset width.
        sections = self.get_sections(line_index)

        # wrapped_section is the text that appears on a single y_offset within
        # the TextArea. It's a potentially wrapped portion of a larger line from
        # the original document.
        target_section = sections[y_offset]

        # Add the offsets from the wrapped sections above this one (from the same raw
        # document line)
        target_section_start = sum(
            len(wrapped_section) for wrapped_section in sections[:y_offset]
        )

        # Get the column index within this wrapped section of the line
        target_column_index = target_section_start + cell_width_to_column_index(
            target_section, x_offset, self._tab_width
        )

        # If we're on the final section of a line, the cursor can legally rest beyond
        # the end by a single cell. Otherwise, we'll need to ensure that we're
        # keeping the cursor within the bounds of the target section.
        if y_offset != len(sections) - 1 and y_offset != -1:
            target_column_index = min(
                target_column_index, target_section_start + len(target_section) - 1
            )

        return target_column_index

    def get_sections(self, line_index: int) -> list[str]:
        """Return the sections for the given line index.

        When wrapping is enabled, a single line in the document can visually span
        multiple lines. The list returned represents that visually (each string in
        the list represents a single section (y-offset) after wrapping happens).

        Args:
            line_index: The index of the line to get sections for.

        Returns:
            The wrapped line as a list of strings.
        """
        line_offsets = self._wrap_offsets[line_index]
        wrapped_lines = Text(self.document[line_index], end="").divide(line_offsets)
        return [line.plain for line in wrapped_lines]

    def get_offsets(self, line_index: int) -> list[int]:
        """Given a line index, get the offsets within that line where wrapping
        should occur for the current document.

        Args:
            line_index: The index of the line within the document.

        Raises:
            ValueError: When `line_index` is out of bounds.

        Returns:
            The offsets within the line where wrapping should occur.
        """
        wrap_offsets = self._wrap_offsets
        out_of_bounds = line_index < 0 or line_index >= len(wrap_offsets)
        if out_of_bounds:
            raise ValueError(
                f"The document line index {line_index!r} is out of bounds. "
                f"The document contains {len(wrap_offsets)!r} lines."
            )
        return wrap_offsets[line_index]

    def get_tab_widths(self, line_index: int) -> list[int]:
        """Return a list of the tab widths for the given line index.

        Args:
            line_index: The index of the line in the document.

        Returns:
            An ordered list of the expanded width of the tabs in the line.
        """
        return self._tab_width_cache[line_index]



from textual.app import App, ComposeResult
from textual.widgets import Checkbox, Footer


class ScrollOffByOne(App):
    def compose(self) -> ComposeResult:
        for number in range(1, 100):
            yield Checkbox(str(number))
        yield Footer()

    def on_mount(self) -> None:
        self.query_one("Screen").scroll_end()


app = ScrollOffByOne()
if __name__ == "__main__":
    app.run()



from textual import events
from textual.app import App


async def test_run_test() -> None:
    """Test the run_test context manager."""
    keys_pressed: list[str] = []

    class TestApp(App[str]):
        def on_key(self, event: events.Key) -> None:
            keys_pressed.append(event.key)

    app = TestApp()
    async with app.run_test() as pilot:
        assert str(pilot) in (
            "<Pilot app=TestApp(title='TestApp', classes={'-dark-mode'}, pseudo_classes={'dark', 'focus'})>",
            "<Pilot app=TestApp(title='TestApp', classes={'-dark-mode'}, pseudo_classes={'focus', 'dark'})>",
        )
        await pilot.press("tab", *"foo")
        await pilot.exit("bar")

    assert app.return_value == "bar"
    assert keys_pressed == ["tab", "f", "o", "o"]



"""
Fuzzy matcher.

This class is used by the [command palette](/guide/command_palette) to match search terms.

"""

from __future__ import annotations

from re import IGNORECASE, compile, escape

import rich.repr
from rich.style import Style
from rich.text import Text

from textual.cache import LRUCache


@rich.repr.auto
class Matcher:
    """A fuzzy matcher."""

    def __init__(
        self,
        query: str,
        *,
        match_style: Style | None = None,
        case_sensitive: bool = False,
    ) -> None:
        """Initialise the fuzzy matching object.

        Args:
            query: A query as typed in by the user.
            match_style: The style to use to highlight matched portions of a string.
            case_sensitive: Should matching be case sensitive?
        """
        self._query = query
        self._match_style = Style(reverse=True) if match_style is None else match_style
        self._query_regex = compile(
            ".*?".join(f"({escape(character)})" for character in query),
            flags=0 if case_sensitive else IGNORECASE,
        )
        self._cache: LRUCache[str, float] = LRUCache(1024 * 4)

    @property
    def query(self) -> str:
        """The query string to look for."""
        return self._query

    @property
    def match_style(self) -> Style:
        """The style that will be used to highlight hits in the matched text."""
        return self._match_style

    @property
    def query_pattern(self) -> str:
        """The regular expression pattern built from the query."""
        return self._query_regex.pattern

    @property
    def case_sensitive(self) -> bool:
        """Is this matcher case sensitive?"""
        return not bool(self._query_regex.flags & IGNORECASE)

    def match(self, candidate: str) -> float:
        """Match the candidate against the query.

        Args:
            candidate: Candidate string to match against the query.

        Returns:
            Strength of the match from 0 to 1.
        """
        cached = self._cache.get(candidate)
        if cached is not None:
            return cached
        match = self._query_regex.search(candidate)
        if match is None:
            score = 0.0
        else:
            assert match.lastindex is not None
            offsets = [
                match.span(group_no)[0] for group_no in range(1, match.lastindex + 1)
            ]
            group_count = 0
            last_offset = -2
            for offset in offsets:
                if offset > last_offset + 1:
                    group_count += 1
                last_offset = offset

            score = 1.0 - ((group_count - 1) / len(candidate))
        self._cache[candidate] = score
        return score

    def highlight(self, candidate: str) -> Text:
        """Highlight the candidate with the fuzzy match.

        Args:
            candidate: The candidate string to match against the query.

        Returns:
            A [rich.text.Text][`Text`] object with highlighted matches.
        """
        match = self._query_regex.search(candidate)
        text = Text.from_markup(candidate)
        if match is None:
            return text
        assert match.lastindex is not None
        if self._query in text.plain:
            # Favor complete matches
            offset = text.plain.index(self._query)
            text.stylize(self._match_style, offset, offset + len(self._query))
        else:
            offsets = [
                match.span(group_no)[0] for group_no in range(1, match.lastindex + 1)
            ]
            for offset in offsets:
                text.stylize(self._match_style, offset, offset + 1)

        return text



from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widgets import Static


class UtilityContainersExample(App):
    CSS_PATH = "utility_containers.tcss"

    def compose(self) -> ComposeResult:
        yield Horizontal(
            Vertical(
                Static("One"),
                Static("Two"),
                classes="column",
            ),
            Vertical(
                Static("Three"),
                Static("Four"),
                classes="column",
            ),
        )


if __name__ == "__main__":
    app = UtilityContainersExample()
    app.run()
